from datetime import datetime

from sqlalchemy import types, Column
from sqlalchemy.orm import declarative_base

Base = declarative_base()


class Run(Base):
    __tablename__ = 'runs'

    id = Column(types.DATE, primary_key=True)
    blob = Column(types.JSON())


class Experiment(Base):
    __tablename__ = 'experiments'

    id = Column(types.INTEGER,
                primary_key=True,
                autoincrement=True)
    executed_at = Column(types.DATE)
    command = Column(types.String)
    git_repo = Column(types.String)
    git_commit_hash = Column(types.String(12))
    metrics = Column(types.JSON)
    username = Column(types.String)
    github_name = Column(types.String)
    github_email = Column(types.String)
    created_at = Column(types.DATE, default=datetime.now)



from pathlib import Path
from typing import Union

from cloudpathlib import AnyPath, S3Path
from cloudpathlib.local import LocalS3Path


def get_path_from_string(path: str) -> Path:
    return AnyPath(path)


def is_local_path(path: Union[str, Path]) -> bool:
    if isinstance(path, str):
        path = get_path_from_string(path)

    return isinstance(path, Path)


def is_s3_path(path: Union[str, Path]) -> bool:
    if isinstance(path, str):
        path = get_path_from_string(path)

    return isinstance(path, S3Path) or isinstance(path, LocalS3Path)



from anytree import Node

from pycognaize.common.enums import IqDocumentKeysEnum


class ClassificationLabels(dict):

    def __init__(self, raw: dict):
        super().__init__()
        self._label_names = []

        if raw is not None:
            self.create_labels(raw)

    def create_labels(self, raw: dict):
        for src_field_id, data in raw.get(
                IqDocumentKeysEnum.field_category.value, {}).items():
            root = Node('root')
            for label_id, label_names in data.items():
                self._label_names.append(label_names[-1])
                parent = Node(label_names[0], parent=root)
                for label_name in label_names[1:]:
                    node = Node(label_name, parent=parent)
                    parent = node
            self.__setitem__(src_field_id, root)

    @property
    def labels(self):
        return self.__dict__

    @property
    def label_names(self):
        return self._label_names



import importlib
from pycognaize.common.decorators import module_not_found
from pycognaize.common.enums import PageLayoutEnum
from pycognaize.document.document import Document
from pycognaize.document.field import Field, TableField


class LangchainLoader:

    """ Convert Pycognaize Document Object to Langchain Document Object """

    INPUT_FIELDS: list[str] = list(i.value for i in PageLayoutEnum)
    OVERLAP = 512
    LIMIT = 2048

    @module_not_found()
    def __init__(self, document: Document) -> None:
        """The constructor takes a cognaize document object as an input.
        The cognaize document object should have page layout and table data in
        document.x, the pythonnames for each field is defined as a class
        attribute.
        """
        transformers = importlib.import_module('transformers')
        self.document = document
        self.tokenizer = transformers.GPT2TokenizerFast.from_pretrained("gpt2")

    @module_not_found()
    def load_and_split(self):
        """
        load and split the Document into separate langchain document objects
        """
        langchain = importlib.import_module('langchain')
        text_blocks_w_metadata: list[list[tuple[dict, str]]] =\
            self._get_as_text(self.document)
        metadata_list, text_list = self._create_text_and_metadata(
            text_blocks_w_metadata,
            document_id=self.get_document_src())
        text_splitter = langchain.text_splitter.RecursiveCharacterTextSplitter(
            chunk_size=self.LIMIT,
            chunk_overlap=self.OVERLAP,
            length_function=self.count_tokens)
        docs = text_splitter.create_documents(texts=text_list,
                                              metadatas=metadata_list)
        return docs

    def get_document_src(self) -> str:
        """Get the SHA of the document"""
        return self.document.metadata['src']

    def count_tokens(self, text: str) -> int:
        """Tokenize the text and count the number of tokens"""
        return len(self.tokenizer.encode(text))

    @staticmethod
    def _get_table_group(current_group: list[tuple[dict, str]],
                         metadata, text_block):
        if current_group and current_group[-1][0]['block'] in \
                (PageLayoutEnum.PAGE_HEADER,
                 PageLayoutEnum.SECTION_HEADER):
            table_group = [current_group[-1], (metadata, text_block)]
            current_group = current_group[:-1]
        else:
            table_group = [(metadata, text_block)]

        return table_group, current_group

    def _get_as_text(self, document: Document) -> list[list[tuple[dict, str]]]:
        """Given a cognaize document object, return a list of strings,
        where each string represents a single chunk/block.
        The tables and text are always in separate groups.
        """
        fields = []
        for pname in self.INPUT_FIELDS:
            if pname not in document.x:
                continue
            block_fields = [(pname, i) for i in document.x[pname]]
            fields.extend(block_fields)
        ordered_blocks_w_metadata: list[
            tuple[dict, str]] = self._order_text_blocks(fields)
        groups = []
        current_group = None
        for block_n, (metadata, text_block) in \
                enumerate(ordered_blocks_w_metadata):
            # First block, create a new group
            if block_n == 0:
                current_group = [(metadata, text_block)]
            # Table, create a separate group
            if metadata['block'] is PageLayoutEnum.TABLE:
                table_group, current_group =\
                    self._get_table_group(current_group, metadata, text_block)
                # finalize the previous group
                groups.append(current_group)
                # finalized the table group
                groups.append(table_group)
                # create an empty group
                current_group = []
            elif metadata['block'] in (PageLayoutEnum.PAGE_HEADER,
                                       PageLayoutEnum.SECTION_HEADER):
                # If all elements are headers in the group, continue with the
                # same group
                if all((
                        g[0]['block']
                        in (PageLayoutEnum.PAGE_HEADER,
                            PageLayoutEnum.SECTION_HEADER)
                        for g in current_group
                )):
                    current_group.append((metadata, text_block))
                elif current_group:
                    groups.append(current_group)
                    current_group = [(metadata, text_block)]
                else:
                    current_group = [(metadata, text_block)]
                # Create a new group
            elif block_n == len(ordered_blocks_w_metadata) - 1 and\
                    current_group:
                groups.append(current_group)
            else:
                current_group.append((metadata, text_block))
        return groups

    @staticmethod
    def _create_text_and_metadata(
        text_blocks_w_metadata: list[list[tuple[dict, str]]],
        document_id: str
    ) -> tuple[list[dict], list[str]]:
        """
        Creates metadata and texts for creating LangChain Document objects
        """
        metadata_list = []
        text_list = []
        for group in text_blocks_w_metadata:
            group_text = ''
            group_metadata = {"pages": set(), "source": []}
            for metadata, text in group:
                tag = metadata['tag']
                group_metadata['pages'].add(tag.page.page_number)
                group_metadata['source'].append((tag.left, tag.top,
                                                 tag.right, tag.bottom,
                                                 tag.page.page_number))
                group_text += f'\n{text}'
            group_metadata['pages'] = list(group_metadata['pages'])
            group_metadata['document'] = document_id
            metadata_list.append(group_metadata)
            text_list.append(group_text)
        return metadata_list, text_list

    @staticmethod
    def _order_text_blocks(fields: list[tuple[str, Field]]) -> \
            list[tuple[dict, str]]:
        """
        Order the fields as they appear in the original document.
        Order by page, top coordinate and left coordinate in that order.
        """
        text_blocks = []
        filtered_fields = []
        for block_type, field in fields:
            if isinstance(field, TableField) and field.tags:
                filtered_fields.append(
                    (
                        {
                            'block': PageLayoutEnum(block_type),
                            'tag': field.tags[0],
                        },
                        field
                    )
                )
            else:
                if field.tags and field.tags[0].raw_value.strip():
                    filtered_fields.append(
                        (
                            {
                                'block': PageLayoutEnum(block_type),
                                'tag': field.tags[0],
                            },
                            field
                        )
                    )
        fields: list[tuple[dict, Field]] = sorted(
            filtered_fields,
            key=lambda x: (x[1].tags[0].page.page_number,
                           x[1].tags[0].top, x[1].tags[0].left))

        for metadata, field in fields:
            if isinstance(field, TableField):
                value: str = field.tags[0].to_string()
            else:
                value = field.value
            text_blocks.append((metadata, value))
        return text_blocks



import fnmatch
import logging
import os
from pathlib import Path, PurePath

import boto3
from botocore.config import Config
from botocore.exceptions import ClientError
from botocore.paginate import Paginator

from ..login import Login


class SnapshotDownloader:
    """
        A class for downloading snapshots from an S3 bucket to a
        local destination.

        Attributes:
            PREFIX (str): The prefix used to indicate S3 paths,
            which is 's3://'.

        Methods:
            download(
                snapshot_path, destination_path, exclude=None,
                continue_token=None):
                Downloads snapshots from an S3 bucket to a local destination.

        Private Methods:
            _get_parts_from_path(path):
                Extracts the bucket name and path from an S3 path.
            _init_s3_objects():
                Initializes the S3 client and resource objects.
            _get_page_iterator(bucket_name, continue_token,
            paginator, path_without_bucket):
                Returns an iterator for paginating through S3 bucket contents.
            _get_relogined_page_iterator(bucket_name,
            next_token, pagination_config, path_without_bucket):
                Returns a relogged-in iterator for paginating through
                 S3 bucket contents.
            _relogin():
                Performs reauthentication for AWS credentials.
            _copy_objects_from_page(bucket_name, exclude,
             page, snapshot_path, destination_path):
                Copies objects from an S3 page to the destination.
            _copy_file_to_dest(s3_object, snapshot_path, destination_path):
                Copies a single S3 object to the destination.
            _write_file(path, file_data):
                Writes file data to a local path.
            _should_exclude(file_path, exclude):
                Determines if a file path should be excluded from copying.
        """

    PREFIX = 's3://'

    def _get_parts_from_path(self, path: str):
        """
        Extracts the bucket name and path from an S3 path.

        Args:
            path (str): The S3 path to extract parts from.

        Returns:
            tuple: A tuple containing the bucket name
            and path without the bucket name.
        """

        if not path:
            return '', ''

        if not path.startswith(self.PREFIX):
            raise ValueError('S3 prefix not in path. '
                             'It must start with s3://.')

        path_without_prefix = PurePath(
            path.split(self.PREFIX, 1)[-1]
        )
        path_parts = path_without_prefix.parts
        bucket_name = path_parts[0]
        path_without_bucket = path_without_prefix.relative_to(bucket_name)
        return bucket_name, path_without_bucket

    def _init_s3_objects(self):
        """
        Initializes the AWS S3 client and resource objects
         for interacting with S3.

        This method sets up the AWS S3 client and resource
         objects using the AWS credentials
        obtained from the `Login` class. It also configures
         the client with retry settings.

        Note:
            This method should be called before performing any
             S3-related operations.

        Raises:
            None
        """

        login = Login()
        session = boto3.session.Session(
            login.aws_access_key,
            login.aws_secret_access_key,
            login.aws_session_token
        )

        client_config = Config(retries={
            'max_attempts': 5,
            'mode': 'standard'
        })

        self._client = session.client('s3', config=client_config)
        self._s3 = session.resource('s3', config=client_config)

    def _get_page_iterator(
            self,
            bucket_name: str,
            continue_token: str,
            paginator: Paginator,
            path_without_bucket: str
    ):
        """
        Returns an iterator for paginating through the
        contents of an S3 bucket.

        Args:
            bucket_name (str): The name of the S3 bucket to paginate.
            continue_token (str): An optional continuation
             token for resuming pagination.
            paginator (botocore.paginate.Paginator): The
             paginator object for list objects.
            path_without_bucket (str): The path within the
            S3 bucket to start pagination from.

        Returns:
            tuple: A tuple containing the page iterator
            and pagination configuration.

        The page iterator allows you to iterate through pages
         of S3 objects within the specified bucket
        and path. The pagination configuration may include
         a continuation token for resuming pagination.

        Raises:
            None
        """

        pagination_config = dict()
        if continue_token:
            pagination_config['StartingToken'] = continue_token
        page_iterator = iter(paginator.paginate(
            Bucket=bucket_name,
            Prefix=str(path_without_bucket),
            PaginationConfig=pagination_config
        ))
        return page_iterator, pagination_config

    def _get_relogined_page_iterator(
            self,
            bucket_name: str,
            next_token: str,
            pagination_config: dict,
            path_without_bucket: str
    ):
        """
        Returns a relogged-in iterator for paginating
        through the contents of an S3 bucket.

        This method is used to obtain an iterator for paginating
         through S3 objects within the specified
        bucket and path, while reauthenticating
        the AWS connection objects.

        Args:
            bucket_name (str): The name of the S3 bucket to paginate.
            next_token (str): An optional continuation token for
             resuming pagination.
            pagination_config (dict): The pagination configuration,
             including the continuation token.
            path_without_bucket (str): The path within the S3
             bucket to start pagination from.

        Returns:
            iter: An iterator for paginating through S3 objects.

        The iterator allows you to iterate through pages
        of S3 objects within the specified bucket
        and path, handling reauthentication when needed.
        The updated pagination configuration may
        contain a new continuation token if reauthentication occurs.

        Raises:
            None
        """
        self._relogin()
        paginator = self._client.get_paginator('list_objects_v2')
        if next_token:
            pagination_config['StartingToken'] = next_token
        return iter(paginator.paginate(
            Bucket=bucket_name,
            Prefix=str(path_without_bucket),
            PaginationConfig=pagination_config
        ))

    def _relogin(self):
        """
        Performs reauthentication to refresh AWS credentials.

        This method reauthenticates to AWS to refresh the
        AWS credentials used by the S3 client and resource
        objects. It utilizes the `Login` class to obtain
         updated AWS credentials.

        Note:
            This method should be called when AWS credentials
             have expired, typically due to an 'ExpiredToken'
            error.

        Raises:
            None
        """
        print('Relogin called')
        login = Login()
        login.relogin()
        self._init_s3_objects()

    def _copy_objects_from_page(
            self,
            bucket_name: str,
            exclude: list,
            include: list,
            page: dict,
            snapshot_path: str,
            destination_path: Path
    ):
        """
        Copies objects from an S3 page to the local destination path.

        This method iterates through the objects in the
         specified S3 page and copies them to the local
        destination path, excluding any objects that
         match patterns in the 'exclude' list.

        Args:
            bucket_name (str): The name of the S3 bucket.
            exclude (list): A list of file
            path patterns to
            exclude from copying.
            include (list): A list of file
            path patterns to
            include even if excluded.
            page (dict): The page of S3 objects to copy.
            snapshot_path (str): The original S3 snapshot path.
            destination_path (Path): The local destination path.

        Returns:
            None

        This method copies each object from the S3 page to the
         destination path, excluding objects based on
        the patterns provided in the 'exclude' list.

        Raises:
            None
        """
        if 'Contents' not in page:
            return

        for obj in page['Contents']:
            s3_object = self._s3.Object(bucket_name, obj['Key'])

            if (self._matches_patterns(s3_object.key, exclude)
                    and not self._matches_patterns(s3_object.key, include)):
                continue

            self._copy_file_to_dest(s3_object, snapshot_path, destination_path)

    def _copy_file_to_dest(
            self,
            s3_object, snapshot_path, destination_path):
        """
        Copies a single S3 object to the local destination path.

        This method copies a single S3 object to the
        specified local destination path. It also handles
        reauthentication if an 'ExpiredToken' error is
        encountered while accessing the S3 object.

        Args:
            s3_object (boto3.resources.factory.s3.Object):
            The S3 object to copy.
            snapshot_path (str): The original S3 snapshot path.
            destination_path (Path): The local destination path.

        Returns:
            None

        This method copies the S3 object to the destination path, and in
         case of an 'ExpiredToken' error,
        it reauthenticates to AWS and retries the copy operation.

        Raises:
            None
        """

        bucket_name = s3_object.bucket_name

        object_full_path = f's3://{str(bucket_name / PurePath(s3_object.key))}'

        print(f'Copying: {object_full_path} to {destination_path}.')

        object_relative_path = (
            PurePath(object_full_path).relative_to(snapshot_path))

        try:
            file_data = s3_object.get()['Body'].read()
        except ClientError as e:
            if e.response['Error']['Code'] == 'ExpiredToken':
                self._relogin()
                s3_object = self._s3.Object(bucket_name, s3_object.key)
                file_data = s3_object.get()['Body'].read()
            else:
                raise

        new_file_path = str(destination_path / object_relative_path)

        self._write_file(new_file_path, file_data)

        print(f'Copied: {object_full_path} to {destination_path}.')

    def _write_file(self, path, file_data):
        """
        Writes file data to a local path.

        This method writes the provided file data to the
        specified local path. It also ensures that the
        necessary directories leading to the path exist,
        and it handles potential conflicts gracefully.

        Args:
            path (str): The local path where the file data should be written.
            file_data (bytes): The binary file data to be written.

        Returns:
            None

        This method creates any missing directories in the path
        and writes the file data to the specified
        location.

        Raises:
            None
        """
        try:
            os.makedirs(os.path.dirname(path), exist_ok=True)
        except FileExistsError:
            logging.info(f'File/folder conflict for '
                         f'{os.path.dirname(path)} path')
            return None

        try:
            with open(path, 'wb') as f:
                f.write(file_data)
        except IsADirectoryError:
            logging.info(f'File/folder conflict for '
                         f'{os.path.dirname(path)} path')

    def _matches_patterns(self, file_path: str,
                          patterns: list):
        """
        Determines whether a file path matches at least one of the patterns.

        This method checks if the provided file path matches any
        of the patterns in the 'patterns' list.
        If a match is found, it returns True, indicating that the
        file is matched; otherwise, it
        returns False.

        Args:
            file_path (str): The file path to be checked for exclusion.
            patterns (list): A list of file path patterns to compare against.

        Returns:
            bool: True if the file path matches one of the patterns,
             False otherwise.

        This method is used to filter out files based on
        specified exclusion or inclusion patterns.

        Raises:
            None
        """
        if not patterns:
            return False

        for pattern in patterns:
            if fnmatch.fnmatch(file_path, pattern):
                return True

        return False

    def download(
            self,
            snapshot_path: str,
            destination_path: str,
            exclude=None,
            include=None,
            continue_token: str = None
    ):
        """
        Downloads snapshots from an S3 bucket to a local destination.

        This method allows you to download snapshots from an S3 bucket to a
         specified local destination
        path. It supports optional exclusion patterns for excluding
        specific files and can continue
        downloading from a specified continuation token.

        Args:
            snapshot_path (str): The S3 path of the snapshot to download.
            destination_path (str): The local destination directory where
                snapshots will be saved.
            exclude (list, optional): A list of file path patterns to
                exclude from copying.
            include (list, optional): A list of file path patterns to
                include even if excluded.
            continue_token (str, optional): An optional continuation token for
                resuming downloads.

        Returns:
            None

        This method initiates the download of snapshots from the specified
         S3 path to the local destination
        path. It handles pagination and reauthentication as needed.
        Progress and status information is
        printed during the download process.

        Note:
            - The 'snapshot_path' should start with 's3://'.
            - The 'destination_path' will be created if it doesn't exist.

        Raises:
            ValueError: If 'snapshot_path' doesn't start with 's3://'.
        """
        if not snapshot_path.startswith('s3://'):
            raise ValueError('Snapshot path should start with "s3://".')

        destination_path = Path(destination_path)

        bucket_name, path_without_bucket = self._get_parts_from_path(
            snapshot_path
        )

        self._init_s3_objects()

        paginator = self._client.get_paginator('list_objects_v2')

        page_iterator, pagination_config = self._get_page_iterator(
            bucket_name, continue_token, paginator,
            path_without_bucket
        )

        next_token = None

        while True:
            try:
                page = next(page_iterator)
                next_token = page.get('NextContinuationToken')

                self._copy_objects_from_page(
                    bucket_name,
                    exclude,
                    include,
                    page,
                    snapshot_path,
                    destination_path
                )
            except ClientError as e:
                if e.response['Error']['Code'] == 'ExpiredToken':
                    page_iterator = self._get_relogined_page_iterator(
                        bucket_name, next_token, pagination_config,
                        path_without_bucket
                    )
                    continue
                raise
            except StopIteration:
                break



import click

from pycognaize import model_registry


@click.group()
def cognaize():
    pass


@cognaize.command()
@click.argument('email', type=click.STRING)
def login(email):
    model_registry.login(email)



import time
from typing import Optional

import requests
from pycognaize import Model

try:
    # noinspection PyPackageRequirements
    from loguru import logger
except ImportError:
    from logging import getLogger
    logger = getLogger(__name__)


class Genie:
    CREATE_TASK_ENDPOINT = "document/modelinput"
    RUN_MODEL_ENDPOINT = "data"
    DIGEST_ENDPOINT = "document/digest"

    def __init__(self, model, x_auth, base_url):

        self.model = model
        self.x_auth = x_auth
        self.base_url = base_url.strip('/') + '/'

    def execute(self, document_id: str, recipe_id: str,
                digest: Optional[bool] = False):
        task_id_response = self.create_and_get_task_id(document_id=document_id,
                                                       recipe_id=recipe_id)
        task_id_response.raise_for_status()
        task_id = task_id_response.json()['taskId']
        logger.info(f"Running the model on task {task_id}")
        start = time.time()
        self.run_model(model=self.model, task_id=task_id)
        duration = time.time() - start
        logger.info(f"Model finished in {duration}")
        if digest:
            logger.info("Digesting model results")
            self.digest_results(task_id=task_id)

    def create_and_get_task_id(self, document_id: str,
                               recipe_id: str) -> requests.Response:
        url = self.base_url + self.CREATE_TASK_ENDPOINT
        payload = {'documentId': document_id, 'recipeId': recipe_id}
        headers = {'x-auth': self.x_auth, 'content-type': "application/json"}
        response = requests.request("POST", url, json=payload,
                                    headers=headers)
        response.raise_for_status()
        return response

    def run_model(self, model: Model, task_id: str) -> requests.Response:
        url = self.base_url + self.RUN_MODEL_ENDPOINT
        response = model.execute_genie_v2(task_id=task_id,
                                          token=self.x_auth,
                                          url=url)
        response.raise_for_status()
        return response

    def digest_results(self, task_id: str) -> requests.Response:
        url = self.base_url + self.DIGEST_ENDPOINT
        payload = {'taskId': task_id}
        headers = {'x-auth': self.x_auth, 'content-type': "application/json"}
        response = requests.request("POST", url, data=payload,
                                    headers=headers)
        response.raise_for_status()
        logger.info(response.text)
        return response



"""Defines LazyDocumentDict, used for lazy loading the documents
in the snapshot
"""
import logging
import os
from collections.abc import Mapping
from typing import Optional

from bson import json_util

from pycognaize.common.enums import StorageEnum
from pycognaize.common.utils import join_path
from pycognaize.document import Document
from pycognaize.file_storage import get_storage
from pycognaize.login import Login


class LazyDocumentDict(Mapping):
    """Contains documents included in the snapshot"""
    document_filename = StorageEnum.doc_file.value

    def __init__(self, doc_path: str,
                 data_path: str):
        login_instance = Login()

        if login_instance.logged_in:
            self._storage_config = {
                'aws_access_key_id': login_instance.aws_access_key,
                'aws_session_token': login_instance.aws_session_token,
                'aws_secret_access_key': login_instance.aws_secret_access_key
            }

        else:
            self._storage_config = None

        self._doc_path = doc_path
        self._data_path = data_path
        storage = get_storage(self._doc_path, config=self._storage_config)

        ids = []

        for directory in storage.list_dir(doc_path, include_files=False):
            if storage.is_file(directory / self.document_filename):
                ids.append(directory.name)

        self._ids = sorted(ids)

    @property
    def doc_path(self) -> str:
        """Path of the document's JSON"""
        return self._doc_path

    @property
    def data_path(self) -> str:
        """Path of the document's OCR and page images"""
        return self._data_path

    def __getitem__(self, doc_id) -> Optional[Document]:
        """The Document object, retrieved from provided path

        Note: Path can be both local and remote
        """

        storage = get_storage(self.doc_path, config=self._storage_config)
        try:
            path = join_path(
                storage.is_s3_path(self.doc_path),
                self.doc_path,
                doc_id,
                f"{self.document_filename}"
            )

            if storage.is_local_path(path):
                with open(path, 'r', encoding='utf8') as f:
                    doc_dict = json_util.loads(f.read())
            else:
                with storage.open(path, 'r') as f:
                    doc_dict = json_util.loads(f.read())
            return Document.from_dict(raw=doc_dict,
                                      data_path=os.path.join(self.data_path,
                                                             doc_id))
        except FileNotFoundError:
            logging.error(f'Document at path {path} is not found.')
        except Exception as e:
            logging.error(f'Failed reading document {doc_id}: {e}')

    def __iter__(self):
        return iter(self._ids)

    def __len__(self):
        return len(self._ids)



from pathlib import Path
from typing import Union, Type

from pycognaize.file_storage.local_storage import LocalStorage
from pycognaize.file_storage.path_type_checker import is_local_path, is_s3_path
from pycognaize.file_storage.s3_storage import S3Storage
from pycognaize.file_storage.storage import Storage

storages = {}


def get_storage_class(key: str) -> Type[Storage]:
    classes = {
        'local': LocalStorage,
        's3': S3Storage
    }

    return classes[key]


def create_storage_instance(key: str, config=None):
    storage_class = get_storage_class(key)
    storage_instance = storage_class(config)
    storages[key] = storage_instance
    return storage_instance


def get_or_create(key: str, config=None):
    if key not in storages:
        return create_storage_instance(key, config)

    if config is None:
        return storages[key]

    return create_storage_instance(key, config)


def get_storage(path: Union[str, Path], config=None) -> Storage:
    if is_local_path(path):
        return get_or_create('local', config)

    if is_s3_path(path):
        return get_or_create('s3', config)

    raise ValueError('Path should be local or S3.')


__all__ = ['get_storage']



from pathlib import Path

import click
import yaml


def login(email):
    credentials_path = Path.home() / '.pycognaize_auth'

    if credentials_path.exists() and not click.confirm(
            'There is an account already logged in do you want to '
            'override the info?'):
        return

    password = click.prompt('Please specify password',
                            type=click.STRING,
                            hide_input=True)

    credentials = {'email': email, 'password': password}

    with open(credentials_path, 'w') as file:
        file.write(yaml.dump(credentials))



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Iterable, Union

from pycognaize.file_storage.path_type_checker \
    import (
        is_local_path,
        is_s3_path,
        get_path_from_string
    )


class Storage(ABC):

    @abstractmethod
    def __init__(self, config=None):
        pass

    @abstractmethod
    def is_dir(self, path: Union[str, Path]) -> bool:
        pass

    @abstractmethod
    def is_file(self, path: Union[str, Path]) -> bool:
        pass

    @abstractmethod
    def _list_dir(self, path: Union[str, Path]) -> Iterable[Path]:
        pass

    @abstractmethod
    def _list_dir_recursive(self, path: Union[str, Path]):
        pass

    def list_dir(
            self,
            path: Union[str, Path],
            recursive=False,
            include_files=True,
            exclude_folders=False
    ) -> Iterable[Path]:
        if isinstance(path, str):
            path = get_path_from_string(path)

        if not path.is_dir():
            raise NotADirectoryError()

        if recursive:
            list_dir = self._list_dir_recursive(path)
        else:
            list_dir = self._list_dir(path)

        for file_path in list_dir:
            if not include_files and file_path.is_file():
                continue
            if exclude_folders and file_path.is_dir():
                continue
            yield file_path

    @abstractmethod
    def open(self, path: Union[str, Path], *args, **kwargs):
        pass

    def is_local_path(self, path: Union[str, Path]) -> bool:
        return is_local_path(path)

    def is_s3_path(self, path: Union[str, Path]) -> bool:
        return is_s3_path(path)

    def get_path_from_string(self, path: str) -> Path:
        return get_path_from_string(path)



import os
from pathlib import Path
from typing import Iterable, Union

from pycognaize.file_storage.storage import Storage


class LocalStorage(Storage):
    def __init__(self, config=None):
        pass

    def is_dir(self, path: Union[str, Path]) -> bool:
        path = Path(path)
        return path.is_dir()

    def is_file(self, path: Union[str, Path]) -> bool:
        path = Path(path)
        return path.is_file()

    def _list_dir_recursive(self, path: Union[str, Path]) -> Iterable[Path]:
        for root, directories, files in os.walk(path):
            for file in files:
                yield Path(os.path.join(root, file))
            for directory in directories:
                yield Path(os.path.join(root, directory))

    def _list_dir(self, path: Union[str, Path]):
        path = Path(path)

        yield from path.iterdir()

    def open(self, path: Union[str, Path], *args, **kwargs):
        path = Path(path)
        return path.open(*args, **kwargs)



import datetime
import json
import os
import subprocess
from pathlib import Path

import click
from sqlalchemy import create_engine, select
from sqlalchemy.orm import Session

from pycognaize.model_registry.db.models import Run, Experiment


class ModelRegistry:

    def __init__(self):
        self.project_path = Path.cwd()

    def submit(self, run_id):
        self._assert_all_changes_committed()

        kedro_metadata = self._get_kedro_metadata(run_id)
        metrics = self._get_kedro_metrics(run_id)
        github_data = self._get_github_data()

        postgres_engine = create_engine(os.getenv('DB_URL'))

        with Session(postgres_engine) as session:
            experiment = Experiment(**kedro_metadata,
                                    metrics=metrics,
                                    **github_data)

            session.add(experiment)
            session.commit()
            run = session.scalars(select(Experiment)).all()

            print(run)

        print(github_data)

    def _assert_all_changes_committed(self):
        try:
            status_process = subprocess.run(['git', 'status', '-s'],
                                            capture_output=True,
                                            check=True,
                                            text=True)

            status_output = status_process.stdout.strip()
        except FileNotFoundError:
            raise click.UsageError('Git should be installed '
                                   'to submit the data.')

        if status_output:
            raise click.UsageError('There are some modifications '
                                   'to the repo. Make sure all changes are'
                                   ' committed.')

    def _get_kedro_metadata(self, run_id):
        db_path = self.project_path / 'data' / 'session_store.db'

        sqlite_engine = create_engine(f'sqlite:///{db_path}')

        with Session(sqlite_engine) as session:
            run = session.scalars(select(Run).where(Run.id == run_id)).first()

            if run is None:
                print('Sorry matching run not found')
                return None

            run.blob = json.loads(run.blob)
            executed_date = datetime.datetime.strptime(run.blob['session_id'],
                                                       '%Y-%m-%dT%H.%M.%S.%fZ')
            return {'executed_at': executed_date,
                    'command': run.blob['cli']['command_path']}

    def _get_kedro_metrics(self, run_id):
        tracking_path = self.project_path / 'data' / '09_tracking'

        metrics = {}

        for metric_dir in tracking_path.iterdir():
            if metric_dir.is_dir():
                metric_dir_name = metric_dir.name

                for run_dir in metric_dir.iterdir():
                    if run_dir.is_dir() and run_dir.name == run_id:
                        file_path = run_dir / metric_dir_name

                        with file_path.open('r') as metric_file:
                            metrics[metric_dir_name] = json.loads(
                                metric_file.read())

        return metrics

    def _get_github_data(self):
        github_data = {}
        try:
            repo_url_command = subprocess.run(
                ['git', 'remote', 'get-url', 'origin'], capture_output=True,
                check=True,
                text=True)
            github_data['git_repo'] = repo_url_command.stdout.strip()

            username_command = subprocess.run(['git', 'config', 'user.name'],
                                              capture_output=True,
                                              check=True,
                                              text=True)
            github_data['github_name'] = username_command.stdout.strip()

            email_command = subprocess.run(['git', 'config', 'user.email'],
                                           capture_output=True,
                                           check=True,
                                           text=True)
            github_data['github_email'] = email_command.stdout.strip()

            commit_hash_command = subprocess.run(['git', 'rev-parse', 'HEAD'],
                                                 capture_output=True,
                                                 check=True,
                                                 text=True)
            github_data['git_commit_hash'] = commit_hash_command.stdout.strip()

        except subprocess.CalledProcessError:
            print('Sorry something went wrong')
        finally:
            return github_data


def submit(run_id):
    registry = ModelRegistry()
    registry.submit(run_id)



from pathlib import Path
from typing import Union, Iterable

from cloudpathlib import CloudPath, S3Client

from pycognaize.file_storage.storage import Storage


class S3Storage(Storage):

    def __init__(self, config=None):
        if config is not None:
            client = S3Client(
                aws_access_key_id=config['aws_access_key_id'],
                aws_session_token=config['aws_session_token'],
                aws_secret_access_key=config['aws_secret_access_key']
            )

            client.set_as_default_client()

    def is_dir(self, path: Union[str, Path]) -> bool:
        path = CloudPath(path)

        return path.is_dir()

    def is_file(self, path: Union[str, Path]) -> bool:
        path = CloudPath(path)

        return path.is_file()

    def _list_dir(self, path: Union[str, Path]) -> Iterable[Path]:
        path = CloudPath(path)

        yield from path.iterdir()

    def _list_dir_recursive(self, path: Union[str, Path]):
        path = CloudPath(path)

        yield from path.rglob('*')

    def open(self, path: Union[str, Path], *args, **kwargs):
        path = CloudPath(path)

        return path.open(*args, **kwargs)
